from PyQt5.QtXml import QDomDocument
from qgis.core import QgsReadWriteContext, QgsLayoutItemMap, \
    QgsLayoutItemLabel, QgsLayoutItemPicture, QgsLayoutItemScaleBar, \
    QgsLayoutItem, QgsMessageLog
from . import paggf_version
import sip


class DBLayoutStorage:
    def __init__(self, storing_field_name='xml_parcel_sketch'):
        self.items_to_store = {}  # dict of item_id and item class
        # self.layer = layer
        self.layer = None
        self.page = None
        self.storing_field_name = storing_field_name

    def store(self, feature, layer, layout):
        doc = QDomDocument()
        root = doc.createElement("data")
        doc.appendChild(root)
        context = QgsReadWriteContext()

        for item in layout.items():
            if isinstance(item, QgsLayoutItem):
                item_id = item.id()

                if not item_id:  # generate an aleatory id
                    item_id = 'id' + item.uuid()[1:-1].replace('-', '')
                doc_elem = doc.createElement(item_id)

                if item.writeXml(doc_elem, doc, context):
                    root.appendChild(doc_elem)
                else:
                    QgsMessageLog.logMessage(
                        'pbm when saving XML (db_store_layout)', 'paggf')

        xml = doc.toString()

        layer.startEditing()
        layer.changeAttributeValue(
            feature.id(),
            layer.fields().indexFromName(self.storing_field_name), xml)

    def load(self, feature, layout):
        xml = feature[self.storing_field_name]

        if xml:
            doc = QDomDocument()
            doc.setContent(xml)
            context = QgsReadWriteContext()

            root = doc.firstChildElement("data")
            root_children = root.childNodes()

            for i in range(root_children.size()):
                container = root_children.item(i)

                if container.isElement():
                    container_elem = container.toElement()
                    container_id = container_elem.tagName()
                    item_raw = layout.itemById(container_id)
                    elem_xml = container_elem.firstChildElement(
                        'LayoutItem')

                    item = None

                    if item_raw:
                        item = sip.cast(
                            item_raw, self.items_to_store[container_id])
                    else:
                        if elem_xml.isElement():
                            elem = elem_xml.toElement()
                            elem_type = int(elem.attribute('type'))
                            if elem_type == 65641:
                                item = QgsLayoutItemLabel(layout)
                            elif elem_type == 65638:  # QgsLayoutItemPage
                                pass

                            if item:
                                layout.addLayoutItem(item)
                            else:
                                QgsMessageLog.logMessage(
                                    'No adding item in layout (id:{}, \
type:{})'.format(container_id, elem_type), 'paggf')
                        else:
                            QgsMessageLog.logMessage(
                                'pbm : elem_xml is not Element', 'paggf')

                    if item:
                        if item.readXml(elem_xml, doc, context):
                            if self.page:
                                item.attemptMove(
                                    item.positionWithUnits(), page=self.page)
                            item.refresh()
                        else:
                            QgsMessageLog.logMessage(
                                '{} : load pbm'.format(container_id), 'paggf')
                    else:
                        QgsMessageLog.logMessage(
                            'No elem to load for (id:{})'.format(container_id),
                            'paggf')
                else:
                    QgsMessageLog.logMessage(
                        'Child is not an elem for {}'.format(container),
                        'paggf')
        else:
            QgsMessageLog.logMessage('PBM : No xml to load', 'paggf')


class Loader:
    ret = {}

    @classmethod
    def get(cls, version):
        if version not in cls.ret:
            items_to_store = {}

            if version == paggf_version.BM:
                items_to_store = {
                    'carto': QgsLayoutItemMap,
                    'cartoText': QgsLayoutItemLabel,
                    'cartoProjection': QgsLayoutItemLabel,
                    'cartoScalebar': QgsLayoutItemScaleBar,
                    'cartoNorthArrow': QgsLayoutItemPicture,
                }

            elif version == paggf_version.CID:
                items_to_store = {
                    'carto': QgsLayoutItemMap,
                    'cartoTextCID': QgsLayoutItemLabel,
                    'cartoText': QgsLayoutItemLabel,
                    'cartoProjection': QgsLayoutItemLabel,
                    'cartoScalebar': QgsLayoutItemScaleBar,
                    'cartoScalebarContainer': QgsLayoutItemLabel,
                    'cartoNorthArrow': QgsLayoutItemPicture,
                }

            cls.ret[version] = DBLayoutStorage()
            cls.ret[version].items_to_store = items_to_store

        return cls.ret[version]
