import os
from PyQt5 import uic
from PyQt5.QtCore import QSettings, Qt
from PyQt5.QtGui import QBrush, QColor, QIcon
from PyQt5.QtWidgets import QWizard, QWizardPage, \
    QFileDialog, QMessageBox, QTableWidgetItem, QHeaderView, \
    QDialog, QListWidgetItem

from qgis.core import QgsFeatureRequest, QgsCoordinateTransform, QgsCoordinateReferenceSystem, \
    QgsGeometry, QgsPoint, QgsMessageLog, QgsVectorLayer, \
    QgsPalLayerSettings, QgsVectorLayerSimpleLabeling, QgsWkbTypes, \
    QgsFeature, QgsFeatureRequest

from .db import get_connection
from .paggf_utils import PAGGFUtils
from .paggf_layers import LoadedLayers
from .maptool_select import MapToolSelectRectangle


PAGE_1_IDC_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_idc_location.ui'))

PAGE_1_BM_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_bm_location.ui'))

PAGE_2_MANUAL_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_manual.ui'))

PAGE_2_FILE_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_file.ui'))

PAGE_3_VERTICES_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'parcel_import_selection.ui'))

NEW_HILL_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'new_hill.ui'))

class NewHillDlg(QDialog, NEW_HILL_W):
    def __init__(self, new_id, newType, parent=None):
        QDialog.__init__(self, parent)
        self.setupUi(self)
        self.setWindowTitle("Ajoutpe d'une nouvelle {}".format(newType))
        
        self.id_edit.setText(new_id)

    def id(self):
        return self.id_edit.text()

    def name(self):
        return self.name_edit.text()

class PageLocationIDC(QWizardPage, PAGE_1_IDC_W):
    def __init__(self, parent=None):
        super(PageLocationIDC, self).__init__(parent)
        self.setupUi(self)

        self.setTitle("Localisation de la parcelle")
        self.setSubTitle("Veuillez renseigner la localisation de la parcelle à importer")

        self.add_province_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "symbologyAdd.svg")))
        self.del_province_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "symbologyRemove.svg")))
        self.add_municipality_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "symbologyAdd.svg")))
        self.del_municipality_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "symbologyRemove.svg")))
        self.add_hill_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "symbologyAdd.svg")))
        self.del_hill_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "symbologyRemove.svg")))
        self.add_sub_hill_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "symbologyAdd.svg")))
        self.del_sub_hill_btn.setIcon(QIcon(os.path.join(os.path.dirname(__file__), "icons", "symbologyRemove.svg")))
        
        self.add_province_btn.clicked.connect(self.on_add_province)
        self.del_province_btn.clicked.connect(self.on_del_province)
        self.add_municipality_btn.clicked.connect(self.on_add_municipality)
        self.del_municipality_btn.clicked.connect(self.on_del_municipality)
        self.add_hill_btn.clicked.connect(self.on_add_hill)
        self.del_hill_btn.clicked.connect(self.on_del_hill)
        self.add_sub_hill_btn.clicked.connect(self.on_add_sub_hill)
        self.del_sub_hill_btn.clicked.connect(self.on_del_sub_hill)

        self.province_field.activated[int].connect(
            self.when_province_selection_changed)
        self.municipality_field.activated[int].connect(
            self.when_municipality_selection_changed)
        self.hill_field.activated[int].connect(
            self.when_hill_selection_changed)

        self.layers = LoadedLayers()
        self.utils = PAGGFUtils(self.layers)

        self.provinces_idx = None
        self.municipalities_idx = None
        self.hills_idx = None
        self.sub_hills_idx = None

    def initializePage(self):

        self.provinces_idx = {}
        self.municipalities_idx = {}
        self.hills_idx = {}
        self.sub_hills_idx = {}

        # populate combos
        province_layer = self.layers.get("ref", "provinces")
        if province_layer:
            idx = 0
            req = QgsFeatureRequest().setSubsetOfAttributes([province_layer.fields().indexFromName(x) for x in ["id", "name"]]).addOrderBy("id")
            for f in province_layer.getFeatures(req):
                self.province_field.addItem(
                    "{} - {}".format(f['id'], f['name']), f['id'])
                self.provinces_idx[f['id']] = idx
                idx += 1

        # select the current municipality, if any:
        current_municipality = self.utils.get_current_municipality()
        if current_municipality:
            id = current_municipality['id']
            self.province_field.setCurrentIndex(self.provinces_idx[id[0:2]])
            self.when_province_selection_changed(self.provinces_idx[id[0:2]])
            self.municipality_field.setCurrentIndex(self.municipalities_idx[id[0:4]])
            self.when_municipality_selection_changed(self.municipalities_idx[id[0:4]])
            self.province_field.setEnabled(False)
            self.municipality_field.setEnabled(False)
        else:
            self.province_field.setCurrentIndex(0)
            self.municipality_field.setCurrentIndex(0)
            self.hill_field.setCurrentIndex(0)
            self.subhill_field.setCurrentIndex(0)
            self.when_province_selection_changed(0)

    def on_add_province(self):
        if '99' in sorted(self.provinces_idx.keys()):
            new_id = "{:02d}".format((len(self.provinces_idx)-1))
        else:
            new_id = "01"

        dlg = NewHillDlg(new_id, "province", parent=self)
        if dlg.exec_() == QDialog.Rejected:
            return

        province_layer = self.layers.get("ref", "provinces")
        province_layer.startEditing()
        f = QgsFeature(province_layer.fields())
        f["id"] = dlg.id()
        f["name"] = dlg.name()
        province_layer.addFeatures([f])
        province_layer.commitChanges()

        # refresh combo boxes
        self.province_field.addItem("{} - {}".format(dlg.id(), dlg.name()), dlg.id())
        last = self.province_field.count()-1
        self.provinces_idx[dlg.id()] = last
        self.province_field.setCurrentIndex(last)
        self.when_province_selection_changed(last)

    def on_del_province(self):
        province_id = self.province_field.currentData()
        r = QMessageBox.warning(None, "Confirmation",
                                "Êtes-vous sûr de vouloir supprimer la province {} ?".format(self.province_field.currentText()),
                                QMessageBox.Yes | QMessageBox.No)
        if r == QMessageBox.No:
            return
        # look for the feature id
        province_layer = self.layers.get("ref", "provinces")
        req = QgsFeatureRequest().setFilterExpression("id='{}'".format(province_id))
        for f in province_layer.getFeatures(req):
            feature_id = f.id()
            break
        province_layer.startEditing()
        if province_layer.deleteFeature(feature_id):
            province_layer.commitChanges()
            # refresh combo boxes
            del self.provinces_idx[province_id]
            self.province_field.removeItem(self.province_field.currentIndex())
            self.when_province_selection_changed(self.province_field.currentIndex())        
        else:
            province_layer.rollBack()            
            QMessageBox.warning(None, "Échec",
                                "La suppression de la province a échouée. Elle contient probablement des communes.")

    def on_add_municipality(self):
        province_id = self.province_field.currentData()
        try:
            last_id = max([k for k in self.municipalities_idx if k.startswith(province_id)])
        except:
            try:
                last_id = province_id + "00"
            except:
                QMessageBox.warning(None, "Attention",
                                "La commune n'hérite d'aucune province mère.\nVeuillez d'abord créer une province.")
                return
        
        new_id = "{:04d}".format(int(last_id)+1)
        dlg = NewHillDlg(new_id, "province", parent=self)
        if dlg.exec_() == QDialog.Rejected:
            return

        municipality_layer = self.layers.get("ref", "municipalities")
        municipality_layer.startEditing()
        f = QgsFeature(municipality_layer.fields())
        f["id"] = dlg.id()
        f["name"] = dlg.name()
        f["id_province"] = f["id"][0:2]
        municipality_layer.addFeatures([f])
        municipality_layer.commitChanges()

        # refresh combo boxes
        self.municipality_field.addItem("{} - {}".format(dlg.id(), dlg.name()), dlg.id())
        last = self.municipality_field.count()-1
        self.municipalities_idx[dlg.id()] = last
        self.municipality_field.setCurrentIndex(last)
        self.when_municipality_selection_changed(last)

    def on_del_municipality(self):
        municipality_id = self.municipality_field.currentData()
        r = QMessageBox.warning(None, "Confirmation",
                                "Êtes-vous sûr de vouloir supprimer la commune {} ?".format(self.municipality_field.currentText()),
                                QMessageBox.Yes | QMessageBox.No)
        if r == QMessageBox.No:
            return
        # look for the feature id
        municipality_layer = self.layers.get("ref", "municipalities")
        req = QgsFeatureRequest().setFilterExpression("id='{}'".format(municipality_id))
        for f in municipality_layer.getFeatures(req):
            feature_id = f.id()
            break
        municipality_layer.startEditing()
        if municipality_layer.deleteFeature(feature_id):
            municipality_layer.commitChanges()
            # refresh combo boxes
            del self.municipalities_idx[municipality_id]
            self.municipality_field.removeItem(self.municipality_field.currentIndex())
            self.when_municipality_selection_changed(self.municipality_field.currentIndex())        
        else:
            municipality_layer.rollBack()            
            QMessageBox.warning(None, "Échec",
                                "La suppression de la commune a échouée. Elle contient probablement des collines.")

    def on_add_hill(self):
        mun_id = self.municipality_field.currentData()
        try:
            last_id = max([k for k in self.hills_idx if k.startswith(mun_id)])
        except:
            try:
                last_id = mun_id + "00"
            except:
                QMessageBox.warning(None, "Attention",
                                "La colline n'hérite d'aucune commune mère.\nVeuillez d'abord créer une commune.")
                return

        new_id = "{:06d}".format(int(last_id)+1)
        dlg = NewHillDlg(new_id, "colline", parent=self)
        if dlg.exec_() == QDialog.Rejected:
            return

        hill_layer = self.layers.get("ref", "hill_districts")
        hill_layer.startEditing()
        f = QgsFeature(hill_layer.fields())
        f["id"] = dlg.id()
        f["name"] = dlg.name()
        f["id_zone"] = f["id"][0:4]+"00"
        f["type"] = 'Hill'
        hill_layer.addFeatures([f])
        hill_layer.commitChanges()

        # refresh combo boxes
        self.hill_field.addItem("{} - {}".format(dlg.id(), dlg.name()), dlg.id())
        last = self.hill_field.count()-1
        self.hills_idx[dlg.id()] = last
        self.hill_field.setCurrentIndex(last)
        self.when_hill_selection_changed(last)

    def on_del_hill(self):
        hill_id = self.hill_field.currentData()
        r = QMessageBox.warning(None, "Confirmation",
                                "Êtes-vous sûr de vouloir supprimer la colline {} ?".format(self.hill_field.currentText()),
                                QMessageBox.Yes | QMessageBox.No)
        if r == QMessageBox.No:
            return
        # look for the feature id
        hill_layer = self.layers.get("ref", "hill_districts")
        req = QgsFeatureRequest().setFilterExpression("id='{}'".format(hill_id))
        for f in hill_layer.getFeatures(req):
            feature_id = f.id()
            break
        hill_layer.startEditing()
        if hill_layer.deleteFeature(feature_id):
            hill_layer.commitChanges()
            # refresh combo boxes
            del self.hills_idx[hill_id]
            self.hill_field.removeItem(self.hill_field.currentIndex())
            self.when_hill_selection_changed(self.hill_field.currentIndex())        
        else:
            hill_layer.rollBack()            
            QMessageBox.warning(None, "Échec",
                                "La suppression de la colline a échouée. Elle contient probablement des sous-collines.")

    def on_add_sub_hill(self):
        hill_id = self.hill_field.currentData()
        try:
            subhills = [k for k in self.sub_hills_idx if k.startswith(hill_id)]
            last_id = max(subhills)
        except:
            try:
                last_id = hill_id + "00"
            except:
                QMessageBox.warning(None, "Attention",
                                "La sous-colline n'hérite d'aucune colline mère.\nVeuillez d'abord créer une colline.")
                return

        new_id = "{:08d}".format(int(last_id)+1)
        dlg = NewHillDlg(new_id, "sous-colline", parent=self)
        if dlg.exec_() == QDialog.Rejected:
            return

        hill_layer = self.layers.get("ref", "sub_hill_avenues")
        hill_layer.startEditing()
        f = QgsFeature(hill_layer.fields())
        f["id"] = dlg.id()
        f["name"] = dlg.name()
        f["id_hill_district"] = f["id"][0:6]
        f["type"] = 'SubHill'
        hill_layer.addFeatures([f])
        hill_layer.commitChanges()

        # refresh combo boxes
        self.subhill_field.addItem(dlg.name(), dlg.id())
        last = self.subhill_field.count()-1
        self.sub_hills_idx[dlg.id()] = last
        self.subhill_field.setCurrentIndex(last)

    def on_del_sub_hill(self):
        sub_hill_id = self.subhill_field.currentData()
        r = QMessageBox.warning(None, "Confirmation",
                                "Êtes-vous sûr de vouloir supprimer la sous-colline {} ?".format(self.subhill_field.currentText()),
                                QMessageBox.Yes | QMessageBox.No)
        if r == QMessageBox.No:
            return
        # look for the feature id
        sub_hill_layer = self.layers.get("ref", "sub_hill_avenues")
        req = QgsFeatureRequest().setFilterExpression("id='{}'".format(sub_hill_id))
        for f in sub_hill_layer.getFeatures(req):
            feature_id = f.id()
            break
        sub_hill_layer.startEditing()
        if sub_hill_layer.deleteFeature(feature_id):
            sub_hill_layer.commitChanges()

            # refresh combo boxes
            del self.sub_hills_idx[sub_hill_id]
            self.subhill_field.removeItem(self.subhill_field.currentIndex())
        else:
            sub_hill_layer.rollBack()            
            QMessageBox.warning(None, "Échec",
                                "La suppression de la sous-colline a échouée. Elle contient probablement des parcelles.")
            

    def when_province_selection_changed(self, index):
        self.municipalities_idx = {}
        mun_layer = self.layers.get("ref", "municipalities")
        self.municipality_field.clear()
        if mun_layer:
            idx = 0
            req = QgsFeatureRequest()\
                  .setFilterExpression("substr(id,1,2)={}".format(self.province_field.currentData()))\
                  .setFlags(QgsFeatureRequest.NoGeometry)
            for f in mun_layer.getFeatures(req):
                self.municipality_field.addItem(
                    "{} - {}".format(f['id'], f['name']), f['id'])
                self.municipalities_idx[f['id']] = idx
                idx += 1
        self.when_municipality_selection_changed(0)

    def when_municipality_selection_changed(self, index):
        self.hills_idx = {}
        hill_layer = self.layers.get("ref", "hill_districts")
        self.hill_field.clear()
        if hill_layer:
            idx = 0
            req = QgsFeatureRequest()\
                  .setFilterExpression("substr(id,1,4)={}".format(self.municipality_field.currentData()))\
                  .setFlags(QgsFeatureRequest.NoGeometry)
            for f in hill_layer.getFeatures(req):
                self.hill_field.addItem(
                    "{} - {}".format(f['id'], f['name']), f['id'])
                self.hills_idx[f['id']] = idx
                idx += 1
        self.when_hill_selection_changed(0)

    def when_hill_selection_changed(self, index):
        """
        Called when the user change the selection of the hill (the selected
        hill in the combo-box hill_field has changed)

        Params
        ------
        int index : new index in the combobox of the selected hill
        """
        self.sub_hills_idx = {}
        hill_id = self.hill_field.itemData(index)
        self.subhill_field.clear()
        sub_hill_layer = self.layers.get("ref", "sub_hill_avenues")
        if sub_hill_layer:
            idx = 0
            req = QgsFeatureRequest()\
                  .setFilterExpression("substr(id,1,6)={}".format(self.hill_field.currentData()))\
                  .setFlags(QgsFeatureRequest.NoGeometry)
            for f in sub_hill_layer.getFeatures(req):
                self.subhill_field.addItem(f['name'], f['id'])
                self.sub_hills_idx[f['id']] = idx
                idx += 1

        # mise à jour numéro parcelle
        conn = get_connection()
        cur = conn.cursor()
        cur.execute("select gis.next_parcel_num('{}')".format(hill_id))
        parcel_number = cur.fetchone()[0]
        self.parcel_number_field.setText(str(parcel_number))

    def subhill(self):
        return self.subhill_field.currentData()

    def parcel_number(self):
        return int(self.parcel_number_field.text())

class PageExistingParcelLocation(QWizardPage, PAGE_1_BM_W):
    def __init__(self, table_name, parent=None, allow_new_parcel=False, other_table_name=None):
        super(PageExistingParcelLocation, self).__init__(parent)
        self.setupUi(self)

        self.setTitle("Identification de la parcelle")
        self.setSubTitle("Veuillez renseigner l'identification de la parcelle")
        self.table_name = table_name
        self.other_table_name = other_table_name
        self.allow_new_parcel = allow_new_parcel
        self.radio_button_new_parcel.setEnabled(allow_new_parcel)

    def initializePage(self):
        conn = get_connection()
        cur = conn.cursor()

        sql = "\
        select id, commission_reference, idc_parcel_code from gis.{} \
        where validity_period is null or is_current(validity_period)".format(self.table_name)
        # exclude parcels from the other table if we are creating a new parcel
        if self.allow_new_parcel:
            sql += "\
        except \
        select id, commission_reference, idc_parcel_code from gis.{} \
        where validity_period is null or is_current(validity_period)".format(self.other_table_name)
        cur.execute(sql)

        self.commission_reference_choice.addItem('---', None)
        self.idc_parcel_code_choice.addItem('---', None)

        selected_id = None
        layers = LoadedLayers()
        layer = layers.get("gis", self.table_name + "_to_valid")
        if layer is not None and layer.selectedFeatures():
            selected_id = layer.selectedFeatures()[0]['idc_parcel_code']

        for cur_row in cur:
            self.commission_reference_choice.addItem(cur_row[1], cur_row[0])

            if cur_row[2]:
                self.idc_parcel_code_choice.addItem(cur_row[2], cur_row[0])
            else:
                self.idc_parcel_code_choice.addItem('sans référence', cur_row[0])
        if selected_id is not None:
            self.idc_parcel_code_choice.setCurrentText(selected_id)

        conn.close()

    def id_parcel(self):
        if self.radio_button_idc_parcel.isChecked():
            id_parcel = self.idc_parcel_code_choice.currentData()
        elif self.radio_button_reference.isChecked():
            id_parcel = self.commission_reference_choice.currentData()
        else: # new_parcel
            id_parcel = None
        return id_parcel

class PageManualEdit(QWizardPage, PAGE_2_MANUAL_W):
    def __init__(self, parent=None):
        super(PageManualEdit, self).__init__(parent)
        self.setupUi(self)

        self.setTitle("Saisie des sommets")
        self.setSubTitle("Veuillez saisir les coordonnées des sommets")

    def initializePage(self):
        self.tableWidget.setRowCount(1)
        self.tableWidget.setColumnCount(3)

        self.tableWidget.setHorizontalHeaderItem(0, QTableWidgetItem('X'))
        self.tableWidget.setHorizontalHeaderItem(1, QTableWidgetItem('Y'))
        self.tableWidget.setHorizontalHeaderItem(2, QTableWidgetItem('Name'))

        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.Stretch)

        self.tableWidget.setItem(0, 0, QTableWidgetItem())
        self.tableWidget.setItem(0, 1, QTableWidgetItem())
        self.tableWidget.setItem(0, 2, QTableWidgetItem())

        self.pushButton.clicked.connect(self.add_row)
        self.tableWidget.cellChanged.connect(self.check_cell)

    def check_cell(self, row, column):
        for row in range(self.tableWidget.rowCount()):
            for column in range(2):
                item = self.tableWidget.item(row, column)
                if item:  # si possede widget
                    if item.text() and not self.tableWidget.locale().toDouble(item.text())[1]:
                        item.setBackground(QBrush(QColor(200, 0, 0)))
                    else:
                        item.setBackground(QBrush())

    def add_row(self):
        row_count = self.tableWidget.rowCount()

        self.tableWidget.setRowCount(self.tableWidget.rowCount() + 1)
        self.tableWidget.setItem(row_count, 0, QTableWidgetItem())
        self.tableWidget.setItem(row_count, 1, QTableWidgetItem())
        self.tableWidget.setItem(row_count, 2, QTableWidgetItem())

    def get_features(self):
        for row in range(self.tableWidget.rowCount()):
            x_locale = self.tableWidget.item(row, 0).text()
            y_locale = self.tableWidget.item(row, 1).text()
            
            # Use locale
            x = self.tableWidget.locale().toDouble(x_locale)
            y = self.tableWidget.locale().toDouble(y_locale)
            
            if x[1] and y[1]:  # avoid empty cell
                name = self.tableWidget.item(row, 2).text()
                geom = QgsGeometry(QgsPoint(float(x[0]), float(y[0])))

                yield ({'name':name}, geom)

    def get_name_field(self):
        return 'name'

class PageFile(QWizardPage, PAGE_2_FILE_W):
    def __init__(self, parent=None):
        super(PageFile, self).__init__(parent)
        self.setupUi(self)

        self.setTitle("Fichier à importer")
        self.setSubTitle("Veuillez sélectionner le fichier de sommets à importer")

        self.browse_btn.clicked.connect(self.on_browse)
        # the file path is mandatory
        self.registerField("file_path*", self.file_edit)

        self._layer = None

    def on_browse(self):
        s = QSettings("PAGGF", "PAGGF")
        last_shp_path = s.value("lastSHPParcelPath", ".")

        if not os.path.exists(last_shp_path):
            last_shp_path = os.getcwd()
        shp_path, _ = QFileDialog.getOpenFileName(
            self, "Sélectionnez le fichier de sommets",
            last_shp_path,
            "*.{}".format(self.file_type_combo.currentText().lower()))

        s.setValue("lastSHPParcelPath", shp_path)
        QgsMessageLog.logMessage(shp_path, 'PAGGF')

        self.file_edit.setText(shp_path)

    def file_path(self):
        return self.file_edit.text()

    def layer(self):
        return self._layer

    def validatePage(self):
        layer_file_path = self.file_path()
        layer_file_path_ext = os.path.splitext(layer_file_path)[1]

        provider_lib = "ogr"  # Default provider lib  for shp
        path = self.file_path()  # Default uri for shp
        if layer_file_path_ext.lower() == '.csv':
            provider_lib = "delimitedtext"
            # make sure the file path works on Windows
            f = self.file_path().replace("\\", "/")
            f = f if f.startswith("/") else "/" + f
            path = "\
file://{}?delimiter={}&crs=epsg:32735&xField={}&yField={}".format(
    f, ";", "X", "Y")
        elif layer_file_path_ext.lower() == '.gpx':
            provider_lib = "gpx"
            path = "{}?type=waypoint".format(self.file_path())

        self._layer = QgsVectorLayer(
            path, "data_import", provider_lib)
        if not self._layer.isValid():
            QMessageBox.warning(None, "Problème de chargement", "Problème lors du chargement de la couche.")
            return False
        if self._layer.geometryType() != QgsWkbTypes.PointGeometry:
            QMessageBox.warning(None, "Couche incompatible", "Ce n'est pas une couche de points, chargement impossible.")
            return False

        return True

class PageVertexSelection(QWizardPage, PAGE_3_VERTICES_W):
    def __init__(self, parent=None):
        super(PageVertexSelection, self).__init__(parent)
        self.setupUi(self)
        self._parent = parent

        self.setTitle("Sélection des sommets et attributs")
        self.setSubTitle("Veuillez sélectionner les sommets à importer et renseigner le nom des champs")

    def initializePage(self):
        self.canvas.setLayers([self._parent.layer()])
        extent = self._parent.layer().extent()
        # add 20% around the extent
        extent.scale(1.2)
        self.canvas.setExtent(extent)

        # by default, select all the features
        self._parent.layer().selectAll()

        mt = MapToolSelectRectangle(self.canvas, [self._parent.layer()])
        self.canvas.setMapTool(mt)

        # re-evaluate isComplete on each selection change
        self._parent.layer().selectionChanged.connect(self.completeChanged)

        # register fields
        self.registerField("name_field", self.name_field, "currentText")

        # populate combos to select fields
        for field in self._parent.layer().fields():
            if not field.isNumeric():
                self.name_field.addItem(field.name())

        self.name_field.currentTextChanged.connect(self.on_name_field_changed)
        self.on_name_field_changed()

    def on_name_field_changed(self):
        # update the labeling options
        label_settings = QgsPalLayerSettings()
        label_settings.fieldName = self.name_field.currentText()
        self._parent.layer().setLabeling(QgsVectorLayerSimpleLabeling(label_settings))
        self.canvas.refreshAllLayers()

    def get_selected(self):
        return self._parent.layer().getSelectedFeatures()

    # iterator that returns (feature, transformed_geometry)
    def get_features(self):
        layer = self._parent.layer()
        srid_layer = layer.crs().postgisSrid()

        transf = None
        if srid_layer != 32735:
            transf = QgsCoordinateTransform(
                QgsCoordinateReferenceSystem(srid_layer),
                QgsCoordinateReferenceSystem(32735))

        for f in layer.getSelectedFeatures():
            # ignore null geometries
            if f.geometry().isNull():
                continue
            if transf:
                geom = QgsGeometry(QgsPoint(transf.transform(f.geometry().asPoint())))
            else:
                geom = f.geometry()

            yield (f, geom)

    def isComplete(self):
        return self._parent.layer().selectedFeatureCount() > 0

    def get_name_field(self):
        return self.name_field.currentText()

class LoadIDCParcelWizard(QWizard):
    def __init__(self, from_file, merge_in_parcel, parent=None):
        super(LoadIDCParcelWizard, self).__init__(parent)
        self.setWindowTitle("Assistant pour l'import de parcelle CID")
        if not merge_in_parcel:
            self.location_page = PageExistingParcelLocation('parcels',
                                                            self,
                                                            allow_new_parcel=True,
                                                            other_table_name='idc_parcels')
            self.location_page.radio_button_new_parcel.setChecked(True)
        else:
            self.location_page = PageExistingParcelLocation('idc_parcels', self)
        self.addPage(self.location_page)
        # new parcel
        self.new_parcel_page = PageLocationIDC(self)
        self.addPage(self.new_parcel_page)
        if from_file:
            self.file_page = PageFile(self)
            self.addPage(self.file_page)
            self.selection_page = PageVertexSelection(self)
            self.addPage(self.selection_page)
        else:
            self.selection_page = PageManualEdit(self)
            self.addPage(self.selection_page)

    def nextId(self):
        # special case for the first page
        if self.currentId() == 0:
            if self.location_page.id_parcel() is None:
                # we need a new parcel
                return 1
            else:
                # ok, it is an existing parcel
                return 2
        else:
            return self.page(self.currentId()).nextId()

    def layer(self):
        return self.file_page.layer()

    def get_features(self):
        return self.selection_page.get_features()
    def id_parcel(self):
        if hasattr(self.location_page, "id_parcel"):
            return self.location_page.id_parcel()
        return None

    def subhill(self):
        return self.new_parcel_page.subhill()
    def parcel_number(self):
        return self.new_parcel_page.parcel_number()
    def name_field(self):
        return self.selection_page.get_name_field()

class LoadBMParcelWizard(QWizard):
    def __init__(self, from_file, merge_in_parcel, parent=None):
        super(LoadBMParcelWizard, self).__init__(parent)
        self.setWindowTitle("Assistant pour l'import de parcelle BM")
        if not merge_in_parcel:
            self.location_page = PageExistingParcelLocation('idc_parcels',
                                                            self,
                                                            allow_new_parcel=True,
                                                            other_table_name='parcels')
        else:
            self.location_page = PageExistingParcelLocation('parcels', self)
        self.addPage(self.location_page)
        # new parcel
        self.new_parcel_page = PageLocationIDC(self)
        self.addPage(self.new_parcel_page)
        if from_file:
            self.file_page = PageFile(self)
            self.addPage(self.file_page)
            self.selection_page = PageVertexSelection(self)
            self.addPage(self.selection_page)
        else:
            self.selection_page = PageManualEdit(self)
            self.addPage(self.selection_page)

    def nextId(self):
        # special case for the first page
        if self.currentId() == 0:
            if self.location_page.id_parcel() is None:
                # we need a new parcel
                return 1
            else:
                # ok, it is an existing parcel
                return 2
        else:
            return self.page(self.currentId()).nextId()

    def layer(self):
        return self.file_page.layer()

    def get_features(self):
        return self.selection_page.get_features()

    def id_parcel(self):
        return self.location_page.id_parcel()

    def subhill(self):
        return self.new_parcel_page.subhill()
    def parcel_number(self):
        return self.new_parcel_page.parcel_number()
    def name_field(self):
        return self.selection_page.get_name_field()


LOAD_POLYGON_FILE_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'load_polygon_file.ui'))
class LoadPolygonFilePage(QWizardPage, LOAD_POLYGON_FILE_W):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self.setTitle("Fichier à charger")
        self.setSubTitle("Veuillez renseigner l'emplacement du fichier à charger")

        self.browse_btn.clicked.connect(self.on_browse)
        # the file path is mandatory
        self.registerField("file_path*", self.file_edit)

        self._layer = None

    def on_browse(self):
        s = QSettings("PAGGF", "PAGGF")
        last_shp_path = s.value("lastSHPParcelPath", ".")

        if not os.path.exists(last_shp_path):
            last_shp_path = os.getcwd()
        shp_path, _ = QFileDialog.getOpenFileName(
            self, "Sélectionnez le fichier de polygones",
            last_shp_path,
            "*.shp")

        s.setValue("lastSHPParcelPath", shp_path)

        self.file_edit.setText(shp_path)

    def file_path(self):
        return self.file_edit.text()

    def layer(self):
        return self._layer

    def validatePage(self):
        self._layer = QgsVectorLayer(
            self.file_path(), "data_import", "ogr")
        if not self._layer.isValid():
            QMessageBox.warning(None, "Problème de chargement", "Problème lors du chargement de la couche.")
            return False
        if self._layer.geometryType() != QgsWkbTypes.PolygonGeometry:
            QMessageBox.warning(None, "Couche incompatible", "Ce n'est pas une couche de polygones, chargement impossible.")
            return False

        return True

LOAD_POLYGON_SELECT_IDS_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'load_polygon_select_ids.ui'))
class LoadPolygonSelectIdsPage(QWizardPage, LOAD_POLYGON_SELECT_IDS_W):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self.setTitle("Parcelles à charger")
        self.setSubTitle("Veuillez sélectionner les parcelles à charger")

        self.parent = parent

    def initializePage(self):
        # populate IDC parcel codes
        for f in self.parent.layer().getFeatures():
            item = QListWidgetItem(f['id'])
            item.setData(Qt.UserRole, f.id())
            self.parcel_list_widget.addItem(item)

    def ids(self):
        return [(x.text(), x.data(Qt.UserRole)) for x in self.parcel_list_widget.selectedItems()]

class LoadIDCPolygonWizard(QWizard):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Assistant pour l'import de parcelle")
        self.file_page = LoadPolygonFilePage(self)
        self.addPage(self.file_page)
        self.selection_page = LoadPolygonSelectIdsPage(self)
        self.addPage(self.selection_page)

    def layer(self):
        return self.file_page.layer()

    def ids_parcel(self):
        return self.selection_page.ids()
