import re
import os
import pathlib
from contextlib import contextmanager

import psycopg2


EXPORT_FILE_EXTENSION = 'data'


_ref_tables = [
    'ref.hill_districts',
    'ref.sub_hill_avenues'
]

_table_order = _ref_tables + [
    'land.polygons',
    'land.vertices',
    'land.polygon_vertices',
    'land.parcels',
    'idc.inventories',
    'idc.inventories_adjacent_people',
    'idc.commission_members',
    'idc.crc_commission_members',
    'idc.witnesses',
    'cadastre.registrations',
    'property.titles',
    'property.surveys',
    'land.parcels_easements',
    'land.parcels_usufructs',
    'land.parcel_buildings',
    'land.parcel_roads',
    'land.parcel_rivers',
]


def parse_filepath(filepath):
    try:
        schema, table = filepath.stem.split('_', 1)
    except ValueError:
        raise RuntimeError('Unknown file name format ({})'.format(filepath.name))
    return schema, table, '{}.{}'.format(schema, table)


def parse_data_row(row):
    row_elements = re.split(r'\t', row.rstrip())
    for i, row_element in enumerate(row_elements):
        if row_element == r'\N':
            row_elements[i] = None
    return row_elements


@contextmanager
def read_only_trigger_disabled(cursor, tablename):
    cursor.execute('select _drop_history_table_trigger(%s)', [tablename])
    yield
    cursor.execute('select _add_history_table_trigger(%s)', [tablename])


def _import_data_file(cursor, file_, schema, table, do_on_conflict):
    for row in file_:
        vals = parse_data_row(row)
        q = '''
            insert into {}.{} values ({})
            on conflict (id) do {}
        '''.format(
                schema,
                table,
                ', '.join(['%s'] * len(vals)),
                do_on_conflict)
        cursor.execute(q, vals)


def import_data_file(cursor, filepath):
    schema, table, table_qualified = parse_filepath(filepath)
    with open(str(filepath)) as file_:
        if table_qualified in _ref_tables:
            _import_data_file(cursor, file_, schema, table, 'nothing')
        else:
            with read_only_trigger_disabled(cursor, table_qualified):
                _import_data_file(cursor, file_, schema, table,
                                  'update set validity_period=excluded.validity_period')


def ordered_files(dirname, pattern):
    def key_fn(filepath):
        table_qualified = parse_filepath(filepath)[2]
        try:
            index = _table_order.index(table_qualified)
        except ValueError:
            err = 'No known table for file {}. The known tables are: {}'.format(
                    filepath.name, ', '.join(_table_order))
            raise RuntimeError(err)
        return index

    return sorted(pathlib.Path(dirname).glob(pattern), key=key_fn)


def import_data_files(cursor, dirname):
    pattern = '*.{}'.format(EXPORT_FILE_EXTENSION)
    for filepath in ordered_files(dirname, pattern):
        import_data_file(cursor, filepath)


def export_data_files(cursor, dirname):
    for relation in _table_order:
        schema, table = relation.split('.')
        fn = os.path.join(dirname, '{}_{}.{}'.format(schema, table, EXPORT_FILE_EXTENSION))
        with open(fn, "w") as fo:
            if relation in _ref_tables:
                sql = 'copy {}.{} to stdout'
            else:
                sql = 'copy (select * from {}.{} where validity_period is not null) to stdout'
            sql = sql.format(schema, table)
            cursor.copy_expert(sql, fo)


def main(argv):
    import argparse

    parser = argparse.ArgumentParser(description='Synchronisation des données')
    parser.add_argument('command', choices=['import', 'export'], help='commande à exécuter')
    parser.add_argument('datadir', help='répertoire contenant les fichiers de donnée')
    parser.add_argument('connstr', help='chaîne de connexion à la base de donnée '
                                        '(exemple: "dbname=paggf user=paggf")')
    args = parser.parse_args(argv)

    datadir = args.datadir
    connstr = args.connstr

    conn = psycopg2.connect(connstr)
    cursor = conn.cursor()

    if args.command == 'import':
        import_data_files(cursor, datadir)
    elif args.command == 'export':
        export_data_files(cursor, datadir)

    conn.commit()

    cursor.close()
    conn.close()


if __name__ == '__main__':
    import sys

    main(sys.argv[1:])
